Merge branch 'master' into next
[paraslash.git] / stat.c
diff --git a/stat.c b/stat.c
index 4ba1b0e8485a8125d2fcdd6353e8508e83417340..03d00d771d622a8fb67ad07a99584ee150550b1f 100644 (file)
--- a/stat.c
+++ b/stat.c
@@ -9,7 +9,7 @@
  *  and para_audiod.
  */
 
-
+#include <regex.h>
 #include <sys/types.h>
 #include <dirent.h>
 
 /** The maximal number of simultaneous connections. */
 #define MAX_STAT_CLIENTS 50
 
+extern char *stat_item_values[NUM_STAT_ITEMS];
+
+/** Flags used for the stat command of para_audiod. */
+enum stat_client_flags {
+       /** Enable parser-friendly output. */
+       SCF_PARSER_FRIENDLY = 1,
+};
+
 /**
  * Describes a status client of para_audiod.
  *
@@ -35,7 +43,9 @@ struct stat_client {
        /** The stat client's file descriptor. */
        int fd;
        /** Bitmask of those status items the client is interested in. */
-       long unsigned item_mask;
+       uint64_t item_mask;
+       /** See \ref stat_client flags. s*/
+       unsigned flags;
        /** Its entry in the list of stat clients. */
        struct list_head node;
 };
@@ -61,6 +71,7 @@ static void dump_stat_client_list(void)
  *
  * \param fd The file descriptor of the client.
  * \param mask Bitfield of status items for this client.
+ * \param parser_friendly Enable parser-friendly output mode.
  *
  * Only those status items having the bit set in \a mask will be
  * sent to the client.
@@ -68,7 +79,7 @@ static void dump_stat_client_list(void)
  * \return Positive value on success, or -E_TOO_MANY_CLIENTS if
  * the number of connected clients exceeds #MAX_STAT_CLIENTS.
  */
-int stat_client_add(int fd, long unsigned mask)
+int stat_client_add(int fd, uint64_t mask, int parser_friendly)
 {
        struct stat_client *new_client;
 
@@ -82,9 +93,11 @@ int stat_client_add(int fd, long unsigned mask)
                initialized = 1;
        }
        PARA_INFO_LOG("adding client on fd %d\n", fd);
-       new_client = para_malloc(sizeof(struct stat_client));
+       new_client = para_calloc(sizeof(struct stat_client));
        new_client->fd = fd;
        new_client->item_mask = mask;
+       if (parser_friendly)
+               new_client->flags = SCF_PARSER_FRIENDLY;
        para_list_add(&new_client->node, &client_list);
        dump_stat_client_list();
        num_clients++;
@@ -93,28 +106,34 @@ int stat_client_add(int fd, long unsigned mask)
 /**
  * Write a message to all connected status clients.
  *
* \param msg A \p NULL terminated buffer.
* \param itemnum The number of the status item of \a msg.
  *
  * On write errors, remove the status client from the client list and close its
  * file descriptor.
  */
-void stat_client_write(const char *msg, int itemnum)
+void stat_client_write_item(int item_num)
 {
        struct stat_client *sc, *tmp;
-       size_t len = strlen(msg);
+       struct para_buffer pb = {.flags = 0};
+       struct para_buffer pfpb = {.flags = PBF_SIZE_PREFIX};
+       const uint64_t one = 1;
 
-       if (!initialized || !len)
+       if (!initialized)
                return;
        list_for_each_entry_safe(sc, tmp, &client_list, node) {
                int fd = sc->fd, ret;
 
-               if (!((1 << itemnum) & sc->item_mask))
+               if (!((one << item_num) & sc->item_mask))
                        continue;
                if (write_ok(fd) > 0) {
-                       ret = write(fd, msg, len);
-       //              PARA_DEBUG_LOG("dumped %s to fd %d, ret = %d\n", msg, fd, ret);
-                       if (ret == len)
+                       struct para_buffer *b =
+                               (sc->flags & SCF_PARSER_FRIENDLY)? &pfpb : &pb;
+                       char *msg = stat_item_values[item_num];
+                       if (!b->buf)
+                               WRITE_STATUS_ITEM(b, item_num, "%s\n",
+                                       msg? msg : "");
+                       ret = write(fd, b->buf, b->offset);
+                       if (ret == b->offset)
                                continue;
                }
                /* write error or fd not ready for writing */
@@ -125,6 +144,8 @@ void stat_client_write(const char *msg, int itemnum)
                free(sc);
                dump_stat_client_list();
        }
+       free(pb.buf);
+       free(pfpb.buf);
 //     if (num_clients)
 //             PARA_DEBUG_LOG("%d client(s)\n", num_clients);
 }
@@ -151,31 +172,68 @@ int stat_item_valid(const char *item)
        return -E_UNKNOWN_STAT_ITEM;
 }
 
+/** The minimal length of a status item buffer. */
+#define MIN_STAT_ITEM_LEN 9 /* 5 + 2 + 2, e.g. '0005 00:\n' */
+
 /**
- * Check if line starts with known status item.
+ * Call a function for each complete status item of a buffer.
+ *
+ * \param item_buf The source buffer.
+ * \param num_bytes The length of \a buf.
+ * \param item_handler Function to call for each complete item.
  *
- * \param line Buffer containing the line.
+ * \return Negative on errors, the number of bytes _not_ passed to \a
+ * item_handler.
  *
- * \return If the beginning of \a line matches any paraslash status item and is
- * followed by a colon, the number of that status item is returned. Otherwise,
- * this function returns \p -E_UNKNOWN_STAT_ITEM.
+ * Status items are expected in the format used by parser-friendly output mode
+ * of the stat command of para_client/para_audioc.
  */
-int stat_line_valid(const char *line)
+int for_each_stat_item(char *item_buf, size_t num_bytes,
+       int (*item_handler)(int, char *))
 {
-       int i;
-       size_t line_len;
-
-       if (!line || !*line)
-               return -E_UNKNOWN_STAT_ITEM;
-       line_len = strlen(line);
-       FOR_EACH_STATUS_ITEM(i) {
-               const char *s = status_item_list[i];
-               size_t item_len = strlen(s);
-
-               if (line_len > item_len && line[item_len] == ':' &&
-                               !strncmp(line, s, item_len))
-                       return i;
+       char *buf = item_buf;
+       int len = num_bytes;
+
+       for (;;) {
+               int i, ret, item_len, item_num = 0;
+               if (len < MIN_STAT_ITEM_LEN)
+                       break;
+               ret = read_size_header(buf);
+               if (ret < 0)
+                       return ret;
+               item_len = ret;
+               if (item_len > len - 5) /* item not complete */
+                       break;
+               for (i = 0; i < 2; i++) {
+                       unsigned char c = buf[5 + i];
+                       item_num <<= 4;
+                       if (c >= '0' && c <= '9') {
+                               item_num += c - '0';
+                               continue;
+                       }
+                       if (c >= 'a' && c <= 'f') {
+                               item_num += c - 'a' + 10;
+                               continue;
+                       }
+                       return -E_STAT_ITEM_PARSE;
+               }
+               if (buf[7] != ':' || buf[5 + item_len - 1] != '\n')
+                       return -E_STAT_ITEM_PARSE;
+               buf[5 + item_len - 1] = '\0';
+               if (item_num >= NUM_STAT_ITEMS)
+                       PARA_WARNING_LOG("unknown status item %d: %s\n",
+                               item_num, buf + 8);
+               else {
+                       ret = item_handler(item_num, buf + 8);
+                       if (ret < 0)
+                               return ret;
+               }
+               buf += 5 + item_len;
+               len -= 5 + item_len;
+               assert(len >= 0 && buf <= item_buf + num_bytes);
        }
-       return -E_UNKNOWN_STAT_ITEM;
+       assert(len >= 0);
+       if (len && len != num_bytes)
+               memmove(item_buf, item_buf + num_bytes - len, len);
+       return len;
 }
-
@@ -93,28 +106,34 @@ int stat_client_add(int fd, long unsigned mask)
  * Write a message to all connected status clients.
  *
- * \param msg A \p NULL terminated buffer.
- * \param itemnum The number of the status item of \a msg.
+ * \param item_num The number of the status item of \a msg.
  *
  * On write errors, remove the status client from the client list and close its